
-- # import nat
-- # import id
-- # import prop

infixr 10 ::

data (×) (A,B:Set) : Set where
  pair : A -> B -> A × B

data Env (A, B : Set) : Set where
  nil  : Env A B
  (::) : A × B -> Env A B -> Env A B

infix 5 ∈

data (∈) {A, B : Set} : A × B -> Env A B -> Set where
  lookupTop (a:A)(b:B)(env:Env A B)	: pair a b ∈ pair a b :: env
  lookupRec (a,c:A)(b,d:B)(env:Env A B) : 
      pair a b ∈ env -> pair a b ∈ pair c d :: env

(∉) (A,B:Set) : A × B -> Env A B -> Set
z ∉ env = ¬ (z ∈ env)

data KeyName : Set where
  keyName : Nat -> KeyName

data ConfLvl : Set where
  secret : ConfLvl
  public : ConfLvl

data IntLvl : Set where
  high : IntLvl
  low  : IntLvl

data BasicOrPrim : Set where
  basic	: BasicOrPrim
  prim  : BasicOrPrim

data BasicOrPrimType : BasicOrPrim -> Set where
  intT  : BasicOrPrimType basic
  encT  : KeyName -> BasicOrPrimType prim -> BasicOrPrimType basic
  primT : BasicOrPrimType basic -> ConfLvl -> IntLvl -> BasicOrPrimType prim

BasicType = BasicOrPrimType basic
PrimType  = BasicOrPrimType prim

VarName = Nat

KeyTypeEnv = Env KeyName ConfLvl
VarTypeEnv = Env VarName PrimType

data Integer : Set where
  intV : Nat -> Integer

data Key : Set where
  keyV : Nat -> Key

KeyEnv = Env KeyName Key
VarEnv = Env VarName Integer

primEnc : KeyName -> Integer -> Integer -> Integer
primDec : KeyName -> Integer -> Integer

-- Typing RUL3Z!!! --------------------------------------------------------

data WFValue : Integer -> PrimType -> Set where
  wfInt (n:Integer)(clvl:ConfLvl)(ilvl:IntLvl) : WFValue n (primT intT clvl ilvl)
  wfEnc (τ:PrimType)(k:KeyName)(clvl:ConfLvl)(ilvl:IntLvl) (n2,n3:Integer)
    : WFValue n3 τ -> WFValue (primEnc k n2 n3) (primT (encT k τ) clvl ilvl)

data WFKeyEnv : KeyEnv -> KeyTypeEnv -> Set where
  wfEmptyKE : WFKeyEnv [] []
  wfKE (ke:KeyEnv)(kte:KeyTypeEnv)(kn:KeyName)(k:Key)(s:ConfLvl) :
    ((k':Key)	  -> pair kn k' ∉ ke) ->
    ((s':ConfLvl) -> pair kn s' ∉ kte) ->
    WFKeyEnv ke kte -> WFKeyEnv (pair kn k :: ke) (pair kn s :: kte)

data WFVarEnv : VarEnv -> VarTypeEnv -> Set where
  wfEmptyVE : WFVarEnv [] []
  wfVE (ve:VarEnv)(vte:VarTypeEnv)(x:VarName)(n:Integer)(t:PrimType) :
    ((n':Integer)  -> pair x n' ∉ ve) ->
    ((t':PrimType) -> pair x t' ∉ vte) ->
    WFValue n t -> WFVarEnv ve vte ->
    WFVarEnv (pair x n :: ve) (pair x t :: vte)

data Expr : Set where
  synConst : Nat -> Expr
  synVar   : VarName -> Expr
  synOp	   : Expr -> Expr -> Expr
  synEnc   : KeyName -> Expr -> Expr
  synDec   : KeyName -> Expr -> Expr

data Cmd : Set where
  synSkip   : Cmd
  synVarAss : VarName -> Expr -> Cmd
  synSeq    : Cmd -> Cmd -> Cmd
  synIf	    : Expr -> Cmd -> Cmd -> Cmd
  synWhile  : Expr -> Cmd -> Cmd

data TypeExpr (kte:KeyTypeEnv)(vte:VarTypeEnv) : Expr -> PrimType -> Set where
  typeInt (n:Nat) : TypeExpr kte vte (synConst n) (primT intT public high)
  typeVar (x:VarName)(t:PrimType) :
      pair x t ∈ vte -> TypeExpr kte vte (synVar x) t

data EvalExpr (ke:KeyEnv)(ve:VarEnv) : Expr -> Integer -> Set where
  evalInt (n:Nat) : EvalExpr ke ve (synConst n) (intV n)
  evalVar (x:VarName)(n:Integer) :
      pair x n ∈ ve -> EvalExpr ke ve (synVar x) n

-- Proofs -----------------------------------------------------------------

presVar (ve:VarEnv)(vte:VarTypeEnv)(x:VarName)(n:Integer)(t:PrimType) :
  WFVarEnv ve vte -> pair x n ∈ ve -> pair x t ∈ vte -> WFValue n t
presVar (wfVE _ _ wf _)   lookupTop	 lookupTop     = wf
presVar (wfVE _ _ _ rec) (lookupRec h0) (lookupRec h1) = presVar rec h0 h1
presVar (wfVE abs _ _ _) (lookupRec h)   lookupTop     = absurd (abs _ h)
presVar (wfVE _ abs _ _)  lookupTop	(lookupRec h)  = absurd (abs _ h)


