-- Proofs about lists and list functions
-- Author :: Ulf Norell
-- Created:: 2004-12-14

-- # import Property
-- # import List
-- # import Monoid
-- # import Bool.Proofs

-- Element relation -------------------------------------------------------

PropElem : (A : Set) -> A -> List A -> Prop
PropElem A y (x::xs) = x==y \/ PropElem A y xs
PropElem A y []      = False

-- Properties about append ------------------------------------------------

--- xs ++ [] == xs
propAppNil : (A:Set) -> (xs:List A) ->  xs++nil == xs
propAppNil A []      = reflexivity []
propAppNil A (x::xs) = fol-plugin(propAppNil A xs)

--- Associativity
propAppAssoc : (A:Set) -> (xs,ys,zs : List A) ->
                  ((xs++ys)++zs) == (xs++(ys++zs))
propAppAssoc A []      ys zs = reflexivity (ys++zs)
propAppAssoc A (x::xs) ys zs = fol-plugin(propAppAssoc A xs ys zs)

-- (List A, [], append) is a monoid
monoidLawsNilAppend : (A:Set) -> MonoidLaws (List A) [] (++)
monoidLawsNilAppend A = struct
  id    = struct {left = reflexivity; right =propAppNil A}
  assoc = propAppAssoc A

monoidNilAppend : (A:Set) -> Monoid (List A)
monoidNilAppend A = struct
  ops    = struct {zero=[]; plus=(++)}
  proofs = monoidLawsNilAppend A

-- Properties about map ---------------------------------------------------
{-
mapCompose : (A, B : Set) -> (f : B -> C) -> (g : A -> B) ->
              (xs : List A) ->
              Eq (List C) (map A C (compose A B C f g) xs)
                          (compose (List A) (List B) (List C)
                                    (map B C f 
-}
-- Properties about reverse -----------------------------------------------

--- rev xs ys = reverse xs ++ ys
propRevReverse : (A : Set) -> (xs,ys : List A) -> 
                    (rev xs ys) == (reverse xs ++ ys)
propRevReverse A []      = reflexivity
propRevReverse A (x::xs) =
    fol-plugin(propAppAssoc A, propRevReverse A xs, (++))

--- reverse (x::xs) == reverse xs ++ [x]
propRevCons : (A : Set) -> (x : A) -> (xs : List A) ->
                (reverse (x::xs)) == (reverse xs ++ [x])
propRevCons A x xs = propRevReverse A xs [x]

--- reverse (xs ++ ys) == reverse ys ++ reverse xs
propRevApp : (A : Set) -> (xs, ys : List A) ->
                (reverse (xs++ys)) == (reverse ys ++ reverse xs)
propRevApp A [] ys      = fol-plugin( propAppNil A )  -- just sym
propRevApp A (x::xs) ys = fol-plugin( propRevReverse A
                                    , propAppAssoc A
                                    , propRevApp A xs ys
                                    )

--- reverse (reverse xs) == xs
propReverseReverse : (A : Set) -> (xs : List A) ->
                      (reverse (reverse xs)) == xs
propReverseReverse A []      = reflexivity  []
propReverseReverse A (x::xs) = fol-plugin( propRevApp A
                                        , propReverseReverse A xs
                                        , propRevReverse A
                                        , rev
                                        )

-- Eq List ----------------------------------------------------------------

eqList : (A : Set) -> Eq A -> List A -> List A -> Bool
eqList A equA []      []      = true
eqList A equA (x::xs) []      = false
eqList A equA []      (y::ys) = false
eqList A equA (x::xs) (y::ys) = equA.eq x y && eqList A equA xs ys

eqListRel : (A : Set) -> Eq A -> Relation (List A)
eqListRel A equA = decidableRelation (eqList A equA)

reflEqList : (A : Set) -> (equA : Eq A) ->
              Reflexive (eqListRel A equA)

symEqList : (A : Set) -> (equA : Eq A) ->
             Symmetric (eqListRel A equA)

transEqList : (A : Set) -> (equA : Eq A) ->
               Transitive (eqListRel A equA)

equList : (A : Set) -> Eq A -> Eq (List A)
equList A equA = struct
  eq     = eqList A equA
  proofs = struct
    refl   = reflEqList A equA
    sym    = symEqList A equA
    trans  = transEqList A equA

-- Ord List ---------------------------------------------------------------

-- not so efficient
leqList : (A : Set) -> Ord A -> List A -> List A -> Bool
leqList A ordA []      ys       = true
leqList A ordA (x::xs) []       = false
leqList A ordA (x::xs) (y::ys)  = less ordA x y ||
                                  equal ordA.equ x y && leqList A ordA xs ys

leqListRel : (A : Set) -> Ord A -> Relation (List A)
leqListRel A ordA = decidableRelation (leqList A ordA)

reflLeqList : (A : Set) -> (ordA : Ord A) ->
               Reflexive' (eqListRel A ordA.equ) (leqListRel A ordA)
reflLeqList A ordA []      []      _ = tt
reflLeqList A ordA (x::xs) []      p = p
reflLeqList A ordA []      (y::ys) _ = tt
reflLeqList A ordA (x::xs) (y::ys) p = fol-plugin( propHoldsAndI
                                                 , propHoldsOrI
                                                 , eq_xy
                                                 , leq_xsys
                                                 )
  where
    eq_xy : Holds (ordA.equ.eq x y)
    eq_xy = fol-plugin( p, propHoldsAndE )

    leq_xsys : leqListRel A ordA xs ys
    leq_xsys = fol-plugin( p, propHoldsAndE, reflLeqList A ordA xs ys )

antiSymLeqList : (A : Set) -> (ordA : Ord A) ->
                  AntiSymmetric' (eqListRel A ordA.equ)
                                 (leqListRel A ordA)
{-
antiSymLeqList A ordA [] [] p q = reflexivity (List A) []
antiSymLeqList A ordA (x::xs) [] p q = fol-plugin(p)
antiSymLeqList A ordA [] (y::ys) p q = fol-plugin(q)
antiSymLeqList A ordA (x::xs) (y::ys) p q =
    fol-plugin( antiSymLeqList A ordA xs ys lem2 lem3
              , lem1
              )
  where
    lem1 : Eq A x y
    lem1 = fol-plugin( p, q, propHoldsOrE, propHoldsAndE, ordA.proofs.antisym )

    lem2 : Holds (leqList A ordA xs ys)
    lem2 = fol-plugin( lem1, p, ordA.proofs.refl
                     , propHoldsOrE, propHoldsAndE, propHoldsNotE
                     )

    lem3 : Holds (leqList A ordA ys xs)
    lem3 = fol-plugin( lem1, q, ordA.proofs.refl
                     , propHoldsOrE, propHoldsAndE, propHoldsNotE
                     )
-}

transLeqList : (A : Set) -> (ordA : Ord A) -> Transitive (leqListRel A ordA)

totalLeqList : (A : Set) -> (ordA : Ord A) -> Total      (leqListRel A ordA)

ordList : (A : Set) -> Ord A -> Ord (List A)
ordList A ordA = struct
  leq = leqList A ordA
  equ = equList A ordA.equ
  proofs = struct
    refl     = reflLeqList A ordA
    antisym  = antiSymLeqList A ordA
    trans    = transLeqList A ordA
    total    = totalLeqList A ordA
