{-# OPTIONS -fglasgow-exts #-}

{-| The FOL plugin.
-}

module Plugins.FOL (plugin) where

import Control.Monad
import Control.Monad.Trans

import Syntax.Position		( Range )
import Syntax.Internal		( Value(VTop), Term )

import TypeChecker		( inferType )
import TypeChecker.Monad	( TC )
import TypeChecker.Monad.Errors	( failure )
import TypeChecker.Values	( valueType )

import Utils.Monads

import Plugins
import Plugins.FOL.Types	( FOLGoal(..) )
import Plugins.FOL.Errors	( noInferenceForFOLError )
import Plugins.FOL.Options
import Plugins.FOL.Translation
import Plugins.FOL.Solve	( solveGoals )
import Plugins.FOL.Errors
import Plugins.FOL.Arities

-- The FOL plugin ---------------------------------------------------------

plugin :: SomePlugin
plugin = mkPlugin $ 
    Plugin  { plName		= "fol"
            , plHelp            = helpFOL
	    , plInit		= \_ -> return []
	    , plFinal		= solveGoals
	    , plCheck		= checkFOL
	    , plInfer		= \_ _ _ -> failure noInferenceForFOLError
	    , plOptions		= options
	    , plDefaultOptions	= defaultOptions
	    }

checkFOL :: FOLOptions -> [FOLGoal] -> [Term] -> Value ->
	    TC (Value, [FOLGoal])
checkFOL opts goals es v =
    do	unless (isProp v) $ failure (notOfTypePropError v $ valueType v)

	((pres,goal),lambdas) <-
	    runM $ (,)  <$> mapM termToFOL es
			<*> valueToFOL v
	let (vs,pre) = unzip pres
	let (lambdas',pre',goal')
		| optBinaryApp opts = (lambdas,pre,goal)
		| otherwise	    = removeBinaryApps (lambdas,pre,goal)
	let name = show $ length goals
	return (VTop v, FOLGoal name (lambdas', pre') goal' vs v : goals)

helpFOL :: String
helpFOL = unlines
  [ "Invokes a automated first-order logic (FOL) prover."
  , "The verbosity levels are interpreted as follows: "
  , "    >= 2: Print goals sent to FOL provers."
  , "    >= 3: Print goals in FOL form and proofs."
  , "    >= 4: Print stderr output received from FOL prover."
  , "    >= 5: Print full FOL prover input and output."
  ]
