
module Plugins.FOL.Types where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad.State
import Text.PrettyPrint

import Syntax.Position
import Syntax.Values
import Syntax.Pretty

import TypeChecker.Context

import Utils.Names
import Utils.Map

import Plugins.FOL.Arguments

-- FOL propositions -------------------------------------------------------

data Formula = Predicate String [FOLTerm]
	     | And Formula Formula
	     | Or Formula Formula
	     | Not Formula
             | Implies Formula Formula
             | Equiv Formula Formula
	     | ForAll String (FOLTerm -> Formula)
	     | Exists String (FOLTerm -> Formula)
	     | TRUE
	     | FALSE

data FOLTerm	= FOLFun String [FOLTerm]
		| FOLVar String

-- | A simple textual representation of counter model
type CounterModel = [String]

implies :: Formula -> Formula -> Formula
implies = Implies

equiv :: Formula -> Formula -> Formula
equiv = Equiv

-- | @bigAnd@ produces a big conjunction.
bigAnd :: [Formula] -> Formula
bigAnd [] = TRUE
bigAnd ps = foldr1 And ps

data Goal = Goal Range Arguments Context [(Value, Type)] Type

-- | The representation of a translated first order goal.
data FOLGoal = FOLGoal Range Arguments ([Formula],[Formula]) Formula [Value] Value
		-- ^ A goal consists of the position from which it was called,
		--   a list of premises (definitions of lambda lifted functions,
		--   and axioms), and the formula to be proved. We also
		--   keep the types for pretty printing reasons.

instance HasRange Goal where
    getRange (Goal r _ _ _ _) = r

instance HasRange FOLGoal where
    getRange (FOLGoal r _ _ _ _ _) = r

