
module Utils.Error where

import Control.Monad.Error
import Text.PrettyPrint

import Syntax.Position
import Utils.Pretty

-- The error message type -------------------------------------------------

data ErrorMsg = EM
    { errWhere	:: Range
    , errWhen	:: Doc
    , errWhat	:: Doc
    , errWhy	:: Doc
    , errCode	:: ErrorCode
    }

data ErrorCode = TypeError | ParseError | OptionError
    deriving (Show, Eq)

instance Error ErrorMsg where
    strMsg s	= noMsg	{ errWhat   = text s }
    noMsg	= EM	{ errWhere  = noRange
			, errWhen   = empty
			, errWhat   = empty
			, errWhy    = empty
			, errCode   = TypeError
			}

instance Show ErrorMsg where
    show = show . pretty

instance Pretty ErrorMsg where
    pretty (EM eWhere eWhen eWhat eWhy eCode) =
	vcat	[ wh
		, text "| When    " <> eWhen <> text ","
		, text "|         " <> eWhat
		, because
		]
	where
	    wh = case show eWhere of
		    ""	-> empty
		    s	-> text s
	    because
		| isEmpty eWhy	= empty
		| otherwise	= text "| because " <> eWhy <> text "."

pwords = map text . words

type ErrorFun = ErrorMsg -> ErrorMsg

